import processing.opengl.*;

public static final float DEGREE = PI/180f;

public static final int VIEWTYPE_2D = 0;
public static final int VIEWTYPE_3D = 1;
public static final int VIEWTYPE_ORTHO = 2;

private Maze _maze;
private MazeUtil _mazeUtil;
private MazeView2D _view2d;
private MazeView3D _view3d;
private int _viewType;

private Slider _sliderW;
private Slider _sliderH;
private Button _button2d;
private Button _button3d;
private Button _buttonOrtho;

private float _sliderWValWas = 0;
private float _sliderWVal = 0;
private float _sliderHValWas = 0;
private float _sliderHVal = 0;


public void setup()
{
  size(800,800, OPENGL);
  
  _view2d = new MazeView2D();
  _view3d = new MazeView3D();

  makeMaze(4,4);
  
  _viewType = VIEWTYPE_2D;
  
  _sliderW = new Slider(width/2-75,height-35, 150,12, 0);
  _sliderW.updateValueOnReleaseOnly(false);
  _sliderH = new Slider(width/2-75,height-15, 150,12, 0);
  _sliderH.updateValueOnReleaseOnly(false);
  _button2d = new Button(15, height-35, 18);
  _button3d = new Button(45, height-35, 18);
  _buttonOrtho = new Button(75, height-35, 18);
}

public void draw()
{
  background( _viewType == VIEWTYPE_2D ? 0 : 48);
  
  doAndDrawUi();
  
  float mazeW = (_maze.getWidth()-1) * _view3d.getUnitSize();
  float mazeH = (_maze.getHeight()-1) * _view3d.getUnitSize();
  
  float boundsW = width;
  float boundsH = height;
  
  float rot;

  switch(_viewType)
  {
    case VIEWTYPE_2D:

        camera(0,0,max(mazeW,mazeH)*1.1, 0,0,0, 0.0, 1.0, 0.0); 
        translate(mazeW / -2f, mazeH / -2f);

        _view2d.drawIt();

      break;

    case VIEWTYPE_3D:
    
        camera(0,0,max(mazeW,mazeH)*1.1, 0,0,0, 0.0, 1.0, 0.0); 
        
        rot = map(mouseX, 0,boundsW, DEGREE*-45f, DEGREE*+45f);
        rotateY(rot);
        rot = map(mouseY, 0,boundsH, DEGREE*-45f, DEGREE*+45f);
        rotateX(rot);
        
        translate(mazeW / -2f, mazeH / -2f);

        _view3d.drawIt();
        
        break;

    case VIEWTYPE_ORTHO:
    
        resetMatrix();
        boundsW *= 1.40;
        boundsH *= 1.40;
        ortho(0, boundsW, 0, boundsH, -999999,+999999);
        
        translate(boundsW/2, -boundsH/2);

        rotateX(DEGREE*45);
        rotateZ(DEGREE*45);
        
        //float f = map(mouseX,0,width, DEGREE*-180,DEGREE*+180);
        //rotateY(f);
        
        translate(mazeW / -2f, mazeH / -2f);

        _view3d.drawIt();

      break;
  }
}

public void keyPressed() 
{
  switch (key)
  {
    case '1':
      _viewType = VIEWTYPE_2D;
      break;
    case '2':
      _viewType = VIEWTYPE_3D;
      break;
    case '3':
      _viewType = VIEWTYPE_ORTHO;
      break;
  }
}

private void doAndDrawUi()
{
  if (_button2d.clickedFlag) {
    _viewType = VIEWTYPE_2D;
    _button2d.clickedFlag = false;
  }
  else if (_button3d.clickedFlag) {
    _viewType = VIEWTYPE_3D;
    _button3d.clickedFlag = false;
  }
  else if (_buttonOrtho.clickedFlag) {
    _viewType = VIEWTYPE_ORTHO;
    _buttonOrtho.clickedFlag = false;
  }
  
  _sliderWValWas = _sliderWVal;
  _sliderWVal = _sliderW.getValue();
  _sliderHValWas = _sliderHVal;
  _sliderHVal = _sliderH.getValue();
  
  if (_sliderWVal != _sliderWValWas || _sliderHVal != _sliderHValWas) { 
    int w = (int)map(_sliderWVal, 0,1, 4, 100);
    int h = (int)map(_sliderHVal, 0,1, 4, 100);
    makeMaze(w,h);
  }
  
  //
  
  resetMatrix();

  perspective();
  camera();
  _sliderW.update();
  _sliderH.update();
  _button2d.update();
  _button3d.update();
  _buttonOrtho.update();

}

private void makeMaze(int wallsW, int wallsH)
{
  _maze = new Maze(wallsW, wallsH);
  new MazeUtil().generate(_maze);
  
  _view2d.setMaze(_maze);
  _view3d.setMaze(_maze);
  
  int len = Math.max(_maze.getWidth()-1, _maze.getHeight()-1);
  float f = (width - 40) / len;
  _view2d.setUnitSize(f);
  _view3d.setUnitSize(f);
  // actually only affects visual size of ortho view
}

