class MazeUtil
{
  public final int DIR_NONE = -1;
  public final int DIR_LEFT = 0;
  public final int DIR_RIGHT = 1;
  public final int DIR_UP = 2;
  public final int DIR_DOWN = 3;

  private Maze _m;
  private int _w;
  private int _h;

  private boolean[][] _visited;
  
  private ArrayList<Integer> _scratch;
  
  private ArrayList<Pos> _crumbs;
  private ArrayList<Pos> _history; // .. all crumbs

  
  public void generate(Maze m)
  {
    // setup...
    
    _m = m;
    _m.fill(true);
    _w = _m.getWidth() - 1;
    _h = _m.getHeight() - 1;
    
    _scratch = new ArrayList();
    _scratch.add(0);
    _scratch.add(1);
    _scratch.add(2);
    _scratch.add(3);
    
    // init cells

    _visited = new boolean[_w][_h];

    for (int y = 0; y < _h; y++) {
      for (int x = 0; x < _w; x++) {
        _visited[x][y] = false;
      }
    }

    _crumbs = new ArrayList();
    _history = new ArrayList();

    //

    int edge = (int)random(4); // DIR_LEFT
    Pos p = getEdgePosAndBreakWall(edge);
    _maze.setStartPos(p);
    _visited[p.x][p.y] = true;
    _crumbs.add(p.clone());
    _history.add(p.clone());

    do
    {
        Pos pos = _crumbs.get(_crumbs.size()-1);
        _crumbs.remove(_crumbs.size()-1);
        // println("removed crumb " + pos.x + "," + pos.y);
        // printPosArrayList(_crumbs);
        makePath(pos);
    }
    while (_crumbs.size() > 0);
    
    doEndPosAndBreakWall();
  }
  
  private void makePath(Pos p)
  {
    // println("makePath " + p.x + " " + p.y);
    
    while(true)
    {
      int dir = getLegalDirection(p);
      
      if (dir == DIR_NONE) 
      {
        return;
      }

      switch(dir)
      {
        case DIR_LEFT:
         _maze.setWallV(p.x, p.y, false); 
          p.x--; 
          break;
        case DIR_RIGHT: 
         _maze.setWallV(p.x+1, p.y, false); 
          p.x++; 
          break;
        case DIR_UP:
         _maze.setWallH(p.y, p.x, false);
          p.y--;
          break;
        case DIR_DOWN:
         _maze.setWallH(p.y+1, p.x, false);
          p.y++; 
          break;
      }
      
      _visited[p.x][p.y] = true;

      p.step = _crumbs.size()+1;
      
      _crumbs.add(p.clone());
      _history.add(p.clone());
      
      // println("moved. pos is now " + p.x + "," + p.y + " step: " + p.step);
      // printPosArrayList(_crumbs);
    }
  }
  
  private void printPosArrayList(ArrayList<Pos> a)
  {
    String s = "";
    for (int i = 0; i < a.size(); i++)
    {
      s += Integer.toString(i) + ": " + Integer.toString(a.get(i).x) + "," + Integer.toString(a.get(i).y) + "(" + Integer.toString(a.get(i).step) +  ") | ";
    }
    println(s);
  }
  
  
  private int getLegalDirection(Pos p)
  {
    int[] dirs = getRandomDirectionsList();
    
    for (int i = 0; i < 4; i++)
    {
      int dir = dirs[i];
      
      switch(dir)
      {
        case DIR_LEFT:
          if (p.x > 0 && ! _visited[p.x-1][p.y]) return dir;
          break;
        case DIR_RIGHT:
          if (p.x < _w-1 && ! _visited[p.x+1][p.y]) return dir;
          break;
        case DIR_UP:
          if (p.y > 0 && ! _visited[p.x][p.y-1]) return dir;
          break;
        case DIR_DOWN:
          if (p.y < _h-1 && ! _visited[p.x][p.y+1]) return dir;
          break;
      }
    }
    
    return DIR_NONE;
  }
  
  private int[] getRandomDirectionsList()
  {
    Collections.shuffle(_scratch);
    int[] a = new int[4];
    for (int i = 0; i < 4; i++)
    {
      a[i] = _scratch.get(i);
    }
    return a;
  }
  
  private Pos getEdgePosAndBreakWall(int edge)
  {
    Pos p = new Pos(0,0, -1);
    int posX, posY;
    
    switch (edge)
    {
      case DIR_LEFT:
        posX = 0;
        posY = (int)random(_maze.getHeight()-1);
        p = new Pos(posX,posY, 0);
        _maze.setWallV(posX, posY, false);
        break;

      case DIR_RIGHT:
        posX = _maze.getWidth() - 2;
        posY = (int)random(_maze.getHeight()-1);
        p = new Pos(posX,posY, 0);
        _maze.setWallV(posX+1, posY, false);
        break;

      case DIR_UP:
        posY = 0;
        posX = (int)random(_maze.getWidth()-1);
        p = new Pos(posX,posY, 0);
        _maze.setWallH(posY, posX, false);
        break;

      case DIR_DOWN:
        posY = _maze.getHeight() - 2;
        posX = (int)random(_maze.getWidth()-1);
        p = new Pos(posX,posY, 0);
        _maze.setWallH(posY+1, posX, false);
        break;
    }

    return p;
  }
  
  private void doEndPosAndBreakWall()
  {
    Collections.sort(_history, new PosStepComparator());
    
    Pos p = new Pos(0,0,0);
    
    while (true)
    {
      p = _history.remove(_history.size()-1);
      
      if (p.x == 0) // like dir_left
      {
        _maze.setWallV(p.x, p.y, false);
        break;
      } 
      else if (p.x == _maze.getWidth()-2) // like dir_right
      {      
        _maze.setWallV(p.x+1, p.y, false);
        break;
      }
      else if (p.y == 0) // like dir_up
      {
        _maze.setWallH(p.y, p.x, false);
        break;
      }
      else if (p.y == _maze.getHeight()-2) // like dir_down
      {
        _maze.setWallH(p.y+1, p.x, false);
        break;
      }
      // else keep going
    }

    _maze.setEndPos(p);
  }
  
  private boolean isFull()
  {
    for (int y = 0; y < _h; y++)
    {
      for (int x = 0; x < _w; x++)
      {
        if (! _visited[x][y]) return false;
      }
    }
    return true;
  }
}

// ------------------

class PosStepComparator implements Comparator
{
  public int compare(Object a, Object b)
  {
    Pos pa = (Pos)a;
    Pos pb = (Pos)b;

    if (pa.step > pb.step) 
      return +1;
    else if (pa.step < pb.step)
      return -1;
    else return 0;
  }
  
  public boolean equals(Object a, Object b)
  {
    Pos pa = (Pos)a;
    Pos pb = (Pos)b;
    return pa.step == pb.step;
  }
}
