import processing.core.*; 
import processing.xml.*; 

import processing.opengl.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class MazeApp extends PApplet {



public static final float DEGREE = PI/180f;

public static final int VIEWTYPE_2D = 0;
public static final int VIEWTYPE_3D = 1;
public static final int VIEWTYPE_ORTHO = 2;

private Maze _maze;
private MazeUtil _mazeUtil;
private MazeView2D _view2d;
private MazeView3D _view3d;
private int _viewType;

private Slider _sliderW;
private Slider _sliderH;
private Button _button2d;
private Button _button3d;
private Button _buttonOrtho;

private float _sliderWValWas = 0;
private float _sliderWVal = 0;
private float _sliderHValWas = 0;
private float _sliderHVal = 0;


public void setup()
{
  size(800,800, OPENGL);
  
  _view2d = new MazeView2D();
  _view3d = new MazeView3D();

  makeMaze(4,4);
  
  _viewType = VIEWTYPE_2D;
  
  _sliderW = new Slider(width/2-75,height-35, 150,12, 0);
  _sliderW.updateValueOnReleaseOnly(false);
  _sliderH = new Slider(width/2-75,height-15, 150,12, 0);
  _sliderH.updateValueOnReleaseOnly(false);
  _button2d = new Button(15, height-35, 18);
  _button3d = new Button(45, height-35, 18);
  _buttonOrtho = new Button(75, height-35, 18);
}

public void draw()
{
  background( _viewType == VIEWTYPE_2D ? 0 : 48);
  
  doAndDrawUi();
  
  float mazeW = (_maze.getWidth()-1) * _view3d.getUnitSize();
  float mazeH = (_maze.getHeight()-1) * _view3d.getUnitSize();
  
  float boundsW = width;
  float boundsH = height;
  
  float rot;

  switch(_viewType)
  {
    case VIEWTYPE_2D:

        camera(0,0,max(mazeW,mazeH)*1.1f, 0,0,0, 0.0f, 1.0f, 0.0f); 
        translate(mazeW / -2f, mazeH / -2f);

        _view2d.drawIt();

      break;

    case VIEWTYPE_3D:
    
        camera(0,0,max(mazeW,mazeH)*1.1f, 0,0,0, 0.0f, 1.0f, 0.0f); 
        
        rot = map(mouseX, 0,boundsW, DEGREE*-45f, DEGREE*+45f);
        rotateY(rot);
        rot = map(mouseY, 0,boundsH, DEGREE*-45f, DEGREE*+45f);
        rotateX(rot);
        
        translate(mazeW / -2f, mazeH / -2f);

        _view3d.drawIt();
        
        break;

    case VIEWTYPE_ORTHO:
    
        resetMatrix();
        boundsW *= 1.40f;
        boundsH *= 1.40f;
        ortho(0, boundsW, 0, boundsH, -999999,+999999);
        
        translate(boundsW/2, -boundsH/2);

        rotateX(DEGREE*45);
        rotateZ(DEGREE*45);
        
        //float f = map(mouseX,0,width, DEGREE*-180,DEGREE*+180);
        //rotateY(f);
        
        translate(mazeW / -2f, mazeH / -2f);

        _view3d.drawIt();

      break;
  }
}

public void keyPressed() 
{
  switch (key)
  {
    case '1':
      _viewType = VIEWTYPE_2D;
      break;
    case '2':
      _viewType = VIEWTYPE_3D;
      break;
    case '3':
      _viewType = VIEWTYPE_ORTHO;
      break;
  }
}

private void doAndDrawUi()
{
  if (_button2d.clickedFlag) {
    _viewType = VIEWTYPE_2D;
    _button2d.clickedFlag = false;
  }
  else if (_button3d.clickedFlag) {
    _viewType = VIEWTYPE_3D;
    _button3d.clickedFlag = false;
  }
  else if (_buttonOrtho.clickedFlag) {
    _viewType = VIEWTYPE_ORTHO;
    _buttonOrtho.clickedFlag = false;
  }
  
  _sliderWValWas = _sliderWVal;
  _sliderWVal = _sliderW.getValue();
  _sliderHValWas = _sliderHVal;
  _sliderHVal = _sliderH.getValue();
  
  if (_sliderWVal != _sliderWValWas || _sliderHVal != _sliderHValWas) { 
    int w = (int)map(_sliderWVal, 0,1, 4, 50);
    int h = (int)map(_sliderHVal, 0,1, 4, 50);
    makeMaze(w,h);
  }
  
  //
  
  resetMatrix();

  perspective();
  camera();
  _sliderW.update();
  _sliderH.update();
  _button2d.update();
  _button3d.update();
  _buttonOrtho.update();

}

private void makeMaze(int wallsW, int wallsH)
{
  _maze = new Maze(wallsW, wallsH);
  new MazeUtil().generate(_maze);
  
  _view2d.setMaze(_maze);
  _view3d.setMaze(_maze);
  
  int len = Math.max(_maze.getWidth()-1, _maze.getHeight()-1);
  float f = (width - 40) / len;
  _view2d.setUnitSize(f);
  _view3d.setUnitSize(f);
  // actually only affects visual size of ortho view
}

class Button
{
  // Dumb button:
  // Raises public flag when clicked.
  // Uses public boolean for on-state

  public boolean clickedFlag;
  public boolean selected;

  int _x;
  int _y;
  int _thumbSize;
  
  float _boxx;
  float _boxy;
  
  boolean _mouseDown;
  boolean _mouseDownWas;
  boolean _isOver;
  

  Button(int $x, int $y, int $thumbSize)
  {
    _x = $x;
    _y = $y;
    _thumbSize = $thumbSize;
  }

  public void update()
  {
    _isOver = (mouseX >= _x && mouseX <= _x + _thumbSize &&
              mouseY >= _y && mouseY <= _y + _thumbSize);
    _mouseDownWas = _mouseDown;
    _mouseDown = mousePressed;
   
    if (_isOver && _mouseDown && !_mouseDownWas) 
      clickedFlag = true;
      // (let client detect flag and then clear it)
      
    draw();
  }
  
  private void draw()
  {
    stroke(96,96,96, 128);
    
    if (selected)
      fill(255,255,255, 192);
    else if (_isOver)
      fill(255,255,255, 128);
    else
      fill(255,255,255, 64);
  
    rect( _x, _y, _thumbSize, _thumbSize );
  }
}
interface IMazeView
{
  public void setMaze(Maze m);
  public float getUnitSize();
  public void setUnitSize(float f);
  public void drawIt();
}
class Maze
{
  public Pos _startPos;
  public Pos _endPos;
  
  private int _width;
  private int _height;
  
  private boolean[][] _wallsV;
  private boolean[][] _wallsH;
  
  
  Maze(int w, int h)
  {
    _width = w;
    _height = h;
    _wallsV = new boolean[_width][_height-1];
    _wallsH = new boolean[_height][_width-1];
    
    fill(false);
  }
  
  public void fill(boolean b)
  {
    int x, y;

    // vert walls
    for (x = 0; x < _width; x++) {
      for (y = 0; y < _height-1; y++) {
        _wallsV[x][y] = b;
      }
    }
    // init horiz walls
    for (y = 0; y < _height; y++) {
      for (x = 0; x < _width-1; x++) {
        _wallsH[y][x] = b;
      }
    }
  }
  
  public void makePerimeter()
  {
    // vert walls
    for (int y = 0; y < _height-1; y++) {
      _wallsV[0][y] = true;
      _wallsV[_width-1][y] = true;
    }
    // horiz walls
    for (int x = 0; x < _width-1; x++) {
      _wallsH[0][x] = true;
      _wallsH[_height-1][x] = true;
    }
  }
  
  public void setStartPos(Pos p)
  {
    _startPos = p;  
  }
  
  public void setEndPos(Pos p)
  {
    _endPos = p;  
  }

  public int getWidth()
  {
    return _width;
  }
  
  public int getHeight()
  {
    return _height;
  }
  
  public boolean getWallV(int x, int y)
  {
    return _wallsV[x][y];
  }

  public void setWallV(int x, int y, boolean b)
  {
    _wallsV[x][y] = b;
  }
  
  public boolean getWallH(int y, int x)
  {
    return _wallsH[y][x];
  }  
  
  public void setWallH(int y, int x, boolean b)
  {
    _wallsH[y][x] = b;
  }
}  

class MazeUtil
{
  public final int DIR_NONE = -1;
  public final int DIR_LEFT = 0;
  public final int DIR_RIGHT = 1;
  public final int DIR_UP = 2;
  public final int DIR_DOWN = 3;

  private Maze _m;
  private int _w;
  private int _h;

  private boolean[][] _visited;
  
  private ArrayList<Integer> _scratch;
  
  private ArrayList<Pos> _crumbs;
  private ArrayList<Pos> _history; // .. all crumbs

  
  public void generate(Maze m)
  {
    // setup...
    
    _m = m;
    _m.fill(true);
    _w = _m.getWidth() - 1;
    _h = _m.getHeight() - 1;
    
    _scratch = new ArrayList();
    _scratch.add(0);
    _scratch.add(1);
    _scratch.add(2);
    _scratch.add(3);
    
    // init cells

    _visited = new boolean[_w][_h];

    for (int y = 0; y < _h; y++) {
      for (int x = 0; x < _w; x++) {
        _visited[x][y] = false;
      }
    }

    _crumbs = new ArrayList();
    _history = new ArrayList();

    //

    int edge = (int)random(4); // DIR_LEFT
    Pos p = getEdgePosAndBreakWall(edge);
    _maze.setStartPos(p);
    _visited[p.x][p.y] = true;
    _crumbs.add(p.clone());
    _history.add(p.clone());

    do
    {
        Pos pos = _crumbs.get(_crumbs.size()-1);
        _crumbs.remove(_crumbs.size()-1);
        // println("removed crumb " + pos.x + "," + pos.y);
        // printPosArrayList(_crumbs);
        makePath(pos);
    }
    while (_crumbs.size() > 0);
    
    doEndPosAndBreakWall();
  }
  
  private void makePath(Pos p)
  {
    // println("makePath " + p.x + " " + p.y);
    
    while(true)
    {
      int dir = getLegalDirection(p);
      
      if (dir == DIR_NONE) 
      {
        return;
      }

      switch(dir)
      {
        case DIR_LEFT:
         _maze.setWallV(p.x, p.y, false); 
          p.x--; 
          break;
        case DIR_RIGHT: 
         _maze.setWallV(p.x+1, p.y, false); 
          p.x++; 
          break;
        case DIR_UP:
         _maze.setWallH(p.y, p.x, false);
          p.y--;
          break;
        case DIR_DOWN:
         _maze.setWallH(p.y+1, p.x, false);
          p.y++; 
          break;
      }
      
      _visited[p.x][p.y] = true;

      p.step = _crumbs.size()+1;
      
      _crumbs.add(p.clone());
      _history.add(p.clone());
      
      // println("moved. pos is now " + p.x + "," + p.y + " step: " + p.step);
      // printPosArrayList(_crumbs);
    }
  }
  
  private void printPosArrayList(ArrayList<Pos> a)
  {
    String s = "";
    for (int i = 0; i < a.size(); i++)
    {
      s += Integer.toString(i) + ": " + Integer.toString(a.get(i).x) + "," + Integer.toString(a.get(i).y) + "(" + Integer.toString(a.get(i).step) +  ") | ";
    }
    println(s);
  }
  
  
  private int getLegalDirection(Pos p)
  {
    int[] dirs = getRandomDirectionsList();
    
    for (int i = 0; i < 4; i++)
    {
      int dir = dirs[i];
      
      switch(dir)
      {
        case DIR_LEFT:
          if (p.x > 0 && ! _visited[p.x-1][p.y]) return dir;
          break;
        case DIR_RIGHT:
          if (p.x < _w-1 && ! _visited[p.x+1][p.y]) return dir;
          break;
        case DIR_UP:
          if (p.y > 0 && ! _visited[p.x][p.y-1]) return dir;
          break;
        case DIR_DOWN:
          if (p.y < _h-1 && ! _visited[p.x][p.y+1]) return dir;
          break;
      }
    }
    
    return DIR_NONE;
  }
  
  private int[] getRandomDirectionsList()
  {
    Collections.shuffle(_scratch);
    int[] a = new int[4];
    for (int i = 0; i < 4; i++)
    {
      a[i] = _scratch.get(i);
    }
    return a;
  }
  
  private Pos getEdgePosAndBreakWall(int edge)
  {
    Pos p = new Pos(0,0, -1);
    int posX, posY;
    
    switch (edge)
    {
      case DIR_LEFT:
        posX = 0;
        posY = (int)random(_maze.getHeight()-1);
        p = new Pos(posX,posY, 0);
        _maze.setWallV(posX, posY, false);
        break;

      case DIR_RIGHT:
        posX = _maze.getWidth() - 2;
        posY = (int)random(_maze.getHeight()-1);
        p = new Pos(posX,posY, 0);
        _maze.setWallV(posX+1, posY, false);
        break;

      case DIR_UP:
        posY = 0;
        posX = (int)random(_maze.getWidth()-1);
        p = new Pos(posX,posY, 0);
        _maze.setWallH(posY, posX, false);
        break;

      case DIR_DOWN:
        posY = _maze.getHeight() - 2;
        posX = (int)random(_maze.getWidth()-1);
        p = new Pos(posX,posY, 0);
        _maze.setWallH(posY+1, posX, false);
        break;
    }

    return p;
  }
  
  private void doEndPosAndBreakWall()
  {
    Collections.sort(_history, new PosStepComparator());
    
    Pos p = new Pos(0,0,0);
    
    while (true)
    {
      p = _history.remove(_history.size()-1);
      
      if (p.x == 0) // like dir_left
      {
        _maze.setWallV(p.x, p.y, false);
        break;
      } 
      else if (p.x == _maze.getWidth()-2) // like dir_right
      {      
        _maze.setWallV(p.x+1, p.y, false);
        break;
      }
      else if (p.y == 0) // like dir_up
      {
        _maze.setWallH(p.y, p.x, false);
        break;
      }
      else if (p.y == _maze.getHeight()-2) // like dir_down
      {
        _maze.setWallH(p.y+1, p.x, false);
        break;
      }
      // else keep going
    }

    _maze.setEndPos(p);
  }
  
  private boolean isFull()
  {
    for (int y = 0; y < _h; y++)
    {
      for (int x = 0; x < _w; x++)
      {
        if (! _visited[x][y]) return false;
      }
    }
    return true;
  }
}

// ------------------

class PosStepComparator implements Comparator
{
  public int compare(Object a, Object b)
  {
    Pos pa = (Pos)a;
    Pos pb = (Pos)b;

    if (pa.step > pb.step) 
      return +1;
    else if (pa.step < pb.step)
      return -1;
    else return 0;
  }
  
  public boolean equals(Object a, Object b)
  {
    Pos pa = (Pos)a;
    Pos pb = (Pos)b;
    return pa.step == pb.step;
  }
}
class MazeView3D implements IMazeView
{
  int _colVert = color(0xff675032);
  int _colHoriz = color(0xffaf9761);
  int _colFloor = color(0xffd9cdb3);

  Maze _maze;
  float _unit = 30;
  
  MazeView3D()
  {
  }
  
  public void setMaze(Maze m)
  {
    _maze = m;
  }
  
  public float getUnitSize()
  {
    return _unit;
  }
  
  public void setUnitSize(float f)
  {
    _unit = f;
  }
  
  public void drawIt()
  {
    noStroke();
    
    // vertical slats

    fill(_colVert);

    for (int x = 0; x < _maze.getWidth(); x++)
    {
      for (int y = 0; y < _maze.getHeight() - 1; y++)
      {
        if (_maze.getWallV(x,y))
        {
          float x1 = (float)x*_unit;
          float y1 = (float)y*_unit;
          float x2 = (float)(x+0)*_unit;
          float y2 = (float)(y+1)*_unit;
          //line(x1,y1, x2,y2);

          pushMatrix();
          translate(x1,y1);
          rotateY(-PI/2f);
          quad(0,0, _unit,0, _unit,_unit, 0,_unit);
          popMatrix();
        }
      }
    }
    
    // horiz slats

    fill(_colHoriz);
    
    for (int y = 0; y < _maze.getHeight(); y++)
    {
      for (int x = 0; x < _maze.getWidth() - 1; x++)
      {
        if (_maze.getWallH(y,x))
        {
          float x1 = (float)x*_unit;
          float y1 = (float)y*_unit;
          float x2 = (float)(x+1)*_unit;
          float y2 = (float)(y+0)*_unit;
          // line(x1,y1, x2,y2);
          
          pushMatrix();
          translate(x1,y1);
          rotateX(+PI/2f);
          quad(0,0, _unit,0, _unit,_unit, 0,_unit);
          popMatrix();
          
        }
      }
    }
    
    // floor
    fill(_colFloor);
    float w = (float)(_maze.getWidth()-1)*_unit;
    float h = (float)(_maze.getHeight()-1)*_unit;
    quad(0f,0f, w,0f, w,h, 0f,h);
  }
}

class Pos
{
  public int x;
  public int y;
  public int step; // number of 'steps' from maze startPos
  
  Pos(int x, int y, int step)
  {
    this.x = x;
    this.y = y;
    this.step = step;
  }
  
  public Pos clone()
  {
    return new Pos(this.x, this.y, this.step);
  }
}

class MazeView2D implements IMazeView
{
  int _strokeColor = color(255);

  Maze _maze;
  float _unit = 30;
  
  MazeView2D()
  {
  }
  
  public void setMaze(Maze m)
  {
    _maze = m;
  }
  
  public float getUnitSize()
  {
    return _unit;
  }
  
  public void setUnitSize(float f)
  {
    _unit = f;
  }
  
  public void drawIt()
  {
    stroke(_strokeColor);
    strokeWeight(1);
    
    // vertical slats
    for (int x = 0; x < _maze.getWidth(); x++)
    {
      for (int y = 0; y < _maze.getHeight() - 1; y++)
      {
        if (_maze.getWallV(x,y))
        {
          line(x*_unit, y*_unit, x*_unit, (y+1)*_unit);
        }
      }
    }
    
    // horiz slats
    for (int y = 0; y < _maze.getHeight(); y++)
    {
      for (int x = 0; x < _maze.getWidth() - 1; x++)
      {
        if (_maze.getWallH(y,x))
        {
          line(x*_unit, y*_unit, (x+1)*_unit, y*_unit);
        }
      }
    }
  }
}
class Slider
{
  float _val;   // .. current value
  float _value; // .. official set value, if that makes sense
  int _x;
  int _y;
  int _width;
  int _thumbSize;
  
  float _boxx;
  float _boxy;
  
  boolean _updateValueOnReleaseOnly = false;
  
  boolean _mouseDown;
  boolean _mouseDownWas;
  boolean _isOver;
  boolean _isDragging;
  int _dragXStart;

  Slider(int $x, int $y, int $width, int $thumbSize, float $value)
  {
    _x = $x;
    _y = $y;
    _width = $width;
    _thumbSize = $thumbSize;

    _val = $value;
    setValue(_val);
  }
  
  public void updateValueOnReleaseOnly(boolean $b)
  {
    _updateValueOnReleaseOnly = $b;
  }

  public float getValue()
  {
    return _value;
  }
  
  public void setValue(float $v)
  {
    _val = $v;
    if (_val > 1) _val = 1;
    if (_val < 0) _val = 0;
    
    _value = _val;
  }

  public void update()
  {
    // Biz logic
    
    _boxx = _x - (_thumbSize/2) + (_width * _val);
    _boxy = _y - (_thumbSize/2);

    _isOver = (mouseX >= _boxx && mouseX <= _boxx + _thumbSize &&
              mouseY >= _boxy && mouseY <= _boxy + _thumbSize);
    _mouseDownWas = _mouseDown;
    _mouseDown = mousePressed;
   
    if (_isOver && _mouseDown && !_mouseDownWas) 
    {
      _isDragging = true;
      _dragXStart = mouseX;
    }
      
    //
    
    if (_isDragging) 
    {
      _val = norm(mouseX,  _x, _x + _width);
      if (_val > 1) _val = 1;
      if (_val < 0) _val = 0;
    }

    if (!_mouseDown && _mouseDownWas && _isDragging) 
    {
      _isDragging = false;
      setValue(_val);
    }
      
    if (_isDragging && ! _updateValueOnReleaseOnly) 
    {
      setValue(_val);
    }
    
    draw();
  }
  
  private void draw()
  {
    stroke(96,96,96, 128);
    line(_x, _y, _x + _width, _y);
    
    stroke(0);
    
    if (_isDragging)
      fill(255,255,255, 192);
    else if (_isOver)
      fill(255,255,255, 128);
    else
      fill(255,255,255, 64);
  
    rect( _boxx, _boxy, _thumbSize, _thumbSize );
  }
}
  static public void main(String args[]) {
    PApplet.main(new String[] { "--bgcolor=#F0F0F0", "MazeApp" });
  }
}
